/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "cpu.h"
#include "hardware.h"
#include "log.h"

#define GDB_PORT 5555

static const char hexchars[] = "0123456789abcdef";
static int gdb_socket = -1, client_fd = -1;

static int hex(char ch) {
  if ((ch >= 'a') && (ch <= 'f'))
    return ch - 'a' + 10;
  if ((ch >= '0') && (ch <= '9'))
    return ch - '0';
  if ((ch >= 'A') && (ch <= 'F'))
    return ch - 'A' + 10;
  return -1;
}

static void hex_byte(char *s, int byte) {
  s[0] = hexchars[(byte >> 4) & 0xf];
  s[1] = hexchars[byte & 0xf];
}

static void hex_word(char *s, uint32 val) {
  int i;
  for (i = 0; i < 4; i++)
    hex_byte(s + i * 2, (val >> (i * 8)) & 0xff);
}

static int get_hex_byte(char *s) {
  return (hex(s[0]) << 4) + hex(s[1]);
}

static uint32 get_hex_word(char *s) {
  int i;
  uint32 r = 0;
  for (i = 3; i >= 0; i--)
    r = (r << 8) + get_hex_byte(s + i * 2);
  return r;
}

static void reply_error(int n, char *reply) {
  reply[0] = 'E';
  hex_byte(reply + 1, n);
  reply[3] = 0;
}

static void reply_signal(int n, char *reply) {
  reply[0] = 'S';
  hex_byte(reply + 1, n);
  reply[3] = 0;
}

static void reply_watch(char *type, char *reply) {
  sprintf(reply, "T%02x%s:%08x;", 5, type, cpu->break_addr);
}

static void reply_ok(char *reply) {
  strcpy(reply, "OK");
}

static int open_gdb_socket(void) {
  struct sockaddr_in name;
  int tmp;

  gdb_socket = socket(PF_INET, SOCK_STREAM, 0);
  if (gdb_socket < 0) {
    error_sys("socket: ");
    return -1;
  }

  tmp = 1;
  setsockopt (gdb_socket, SOL_SOCKET, SO_REUSEADDR, &tmp,
	      sizeof (tmp));

  name.sin_family = AF_INET;
  name.sin_port = htons(GDB_PORT);
  name.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(gdb_socket, (struct sockaddr *)&name, sizeof(name)) < 0) {
    error_sys("bind: ");
    close(gdb_socket);
    return -1;
  }

  if (listen(gdb_socket, 1) < 0) {
    error_sys("listen: ");
    close(gdb_socket);
    return -1;
  }

  return 0;
}

static void close_gdb_socket(void) {
  if (gdb_socket >= 0) {
    close(gdb_socket);
    gdb_socket = -1;
  }
}

static int wait_for_client(void) {
  struct sockaddr_in name;
  int fd, size;

  size = sizeof(name);
  client_fd = accept(gdb_socket, (struct sockaddr *)&name, &size);

  return client_fd;
}

static void close_client_socket(void) {
  if (client_fd >= 0) {
    close(client_fd);
    client_fd = -1;
  }
}

static int socket_read(int fd, unsigned char *buf, int len) {
  int r;
  int done = 0;
  while (done < len) {
    r = read(fd, buf + done, len - done);
    if (r == -1) {
      if (errno != EINTR && errno != EAGAIN)
	return -1;
    } else
      done += r;
  }
  return 0;
}

static int socket_get_byte(int fd) {
  unsigned char buf;
  if (socket_read(fd, &buf, 1) < 0)
    return -1;
  else
    return buf;
}

static int socket_write(int fd, unsigned char *buf, int len) {
  int r;
  int done = 0;
  while (done < len) {
    r = write(fd, buf + done, len - done);
    if (r == -1) {
      if (errno != EINTR && errno != EAGAIN)
	return -1;
    } else
      done += r;
  }
  return 0;
}

static int socket_put_byte(int fd, int byte) {
  unsigned char buf = byte;
  return socket_write(fd, &buf, 1);
}

static int socket_try_get_byte(int fd) {
  unsigned char buf;
  int oldflags, newflags;
  int r;

  oldflags = fcntl(fd, F_GETFL, 0);
  if (oldflags == -1)
    return -1;
  newflags = oldflags | O_NONBLOCK;
  if (oldflags != newflags) {
    r = fcntl(fd, F_SETFL, newflags);
    if (r == -1)
      return -1;
  }

  r = read(fd, &buf, 1);

  if (oldflags != newflags)
    fcntl(fd, F_SETFL, oldflags);

  if (r != 1)
    return -1;
  else
    return buf;
}

static int get_packet(char *buf, int len) {
  int count, checksum;
  int ch;

  while (1) {
    do {
      ch = socket_get_byte(client_fd);
      if (ch == -1) return -1;
    } while (ch != '$');

    checksum = 0;
    count = 0;
    while (count < len) {
      ch = socket_get_byte(client_fd);
      if (ch == '$') {
	checksum = 0;
	count = 0;
      } else if (ch == '#')
	break;
      else if (ch == -1)
	return -1;
      else {
	checksum += ch;
	buf[count] = ch;
	count++;
      }
    }
    buf[count] = 0;

    if (ch == '#') {
      int rchksum;

      ch = socket_get_byte(client_fd);
      rchksum = hex(ch) << 4;
      ch = socket_get_byte(client_fd);
      rchksum += hex(ch);

      if ((checksum & 0xff) != rchksum)
	socket_put_byte(client_fd, '-');
      else {
	socket_put_byte(client_fd, '+');
	return count;
      }
    }
  }
}

static int put_packet(char *buf) {
  int i, checksum;
  int ch;

  do {
    socket_put_byte(client_fd, '$');
    checksum = 0;
    i = 0;

    while (ch = buf[i]) {
      socket_put_byte(client_fd, ch);
      checksum += ch;
      i++;
    }

    socket_put_byte(client_fd, '#');
    socket_put_byte(client_fd, hexchars[(checksum >> 4) & 0xf]);
    socket_put_byte(client_fd, hexchars[checksum & 0xf]);

    ch = socket_get_byte(client_fd);

  } while (ch != '+');
}

static void g_reply(char *buf) {
  int i;
  char *p;

  p = buf;
  for (i = 0; i < 16; i++) {
    hex_word(p, cpu->reg[i]);
    p += 8;
  }

  for (i = 0; i < 8; i++) {
    memset(p, '0', 16);
    p += 16;
  }

  hex_word(p, 0);
  p += 8;
  hex_word(p, cpu->cpsr);
  p[8] = 0;
}

static void cmd_set_registers(char *args, char *reply) {
  char *p;
  int i;

  if (strlen(args) < 16 * 8 + 8 * 16 + 2 * 8) {
    reply_error(0, reply);
    return;
  }

  p = args;
  for (i = 0; i < 16; i++) {
    cpu->reg[i] = get_hex_word(p);
    p += 8;
  }

  p += 8 * 16 + 8;

  cpu->cpsr = get_hex_word(p);

  reply_ok(reply);
}

static cmd_get_memory(char *args, char *reply) {
  uint32 addr, len;
  int i;

  if (sscanf(args, "%x,%x", &addr, &len) != 2) {
    reply_error(0, reply);
    return;
  }

  if (len > 8000) {
    reply_error(1, reply);
    return;
  }
  
  for (i = 0; i < len; i++)
    hex_byte(reply + i * 2, memory_read(cpu->mem, addr + i));
  reply[len * 2] = 0;
}

static cmd_put_memory(char *args, char *reply) {
  uint32 addr, len;
  int i, pos;

  pos = -1;
  sscanf(args, "%x,%x:%n", &addr, &len, &pos);
  if (pos == -1) {
    reply_error(0, reply);
    return;
  }

  for (i = 0; i < len; i++)
    memory_write(cpu->mem, addr + i,
		 get_hex_byte(args + pos + i * 2));

  reply_ok(reply);
}

static void parse_continue_args(char *args) {
  int sig;
  uint32 addr;

  if (sscanf(args, "%x;%x", &sig, &addr) == 2) {
    cpu->reg[15] = addr;
  } else if (sscanf(args, "%x", &addr) == 1) {
    cpu->reg[15] = addr;
  }
}

static int gdb_go_handler(void) {
  int r;

  r = socket_try_get_byte(client_fd);
  return (r == 3) ? CPU_STOP_TERMINATE : 0;
}

static int cmd_go(int n, char *args, char *reply) {
  int r;
  parse_continue_args(args);

  r = go(n, gdb_go_handler, 1000);

  switch (r) {
  case CPU_STOP_QUIT:
    strcpy(reply, "W00");
    return 1;
  case CPU_STOP_TERMINATE:
    reply_signal(2, reply);
    break;
  case CPU_STOP_ILLEGAL:
    reply_signal(4, reply);
    break;
  case CPU_STOP_RWATCH:
    reply_watch("rwatch", reply);
    break;
  case CPU_STOP_WWATCH:
    reply_watch("watch", reply);
    break;
  case CPU_STOP_AWATCH:
    reply_watch("awatch", reply);
    break;
  default:
    reply_signal(5, reply);
    break;
  }
  return 0;
}

static void do_set_break(int type, uint32 addr, char *reply) {
  if (cpu_insert_break(cpu, type, addr) < 0) {
    reply_error(1, reply);
    return;
  } else
    reply_ok(reply);
}

static void cmd_set_breakpoint(char *args, char *reply) {
  int type, len;
  uint32 addr;

  if (sscanf(args, "%d,%x,%x", &type, &addr, &len) != 3) {
    reply_error(0, reply);
    return;
  }

  switch (type) {
  case 1:
    do_set_break(CPU_BRK_BREAK, addr, reply);
    reply_ok(reply);
    break;

  case 2:
    do_set_break(CPU_BRK_WWATCH, addr, reply);
    reply_ok(reply);
    break;

  case 3:
    do_set_break(CPU_BRK_RWATCH, addr, reply);
    reply_ok(reply);
    break;

  case 4:
    do_set_break(CPU_BRK_AWATCH, addr, reply);
    reply_ok(reply);
    break;

  default:
    reply[0] = 0;
    break;
  }
}

static void cmd_remove_breakpoint(char *args, char *reply) {
  int type, len;
  uint32 addr;

  if (sscanf(args, "%d,%x,%x", &type, &addr, &len) != 3) {
    reply_error(0, reply);
    return;
  }

  switch (type) {
  case 1:
    cpu_remove_break(cpu, CPU_BRK_BREAK, addr);
    reply_ok(reply);
    break;

  case 2:
    cpu_remove_break(cpu, CPU_BRK_WWATCH, addr);
    reply_ok(reply);
    break;

  case 3:
    cpu_remove_break(cpu, CPU_BRK_RWATCH, addr);
    reply_ok(reply);
    break;

  case 4:
    cpu_remove_break(cpu, CPU_BRK_AWATCH, addr);
    reply_ok(reply);
    break;

  default:
    reply[0] = 0;
    break;
  }
}

void gdb_target(void) {
  char packet_buf[16384];
  char reply_buf[16384];
  int quit = 0, no_reply;

  tick = 0;

  if (open_gdb_socket() < 0)
    return;

  while (!quit) {
    if (wait_for_client() < 0)
      break;

    while (!quit) {
      if (get_packet(packet_buf, sizeof(packet_buf) - 1) < 0) {
	close_client_socket();
	break;
      }

      logmsg(LOG_GDB, 5, "gdb: received %s\n", packet_buf);
      no_reply = 0;
      switch (packet_buf[0]) {
      case '?':
	reply_buf[0] = 'S';
	reply_buf[1] = hexchars[0];
	reply_buf[2] = hexchars[5];
	reply_buf[3] = 0;
	break;

      case 'g':
	g_reply(reply_buf);
	break;

      case 'G':
	cmd_set_registers(packet_buf + 1, reply_buf);
	break;

      case 'm':
	cmd_get_memory(packet_buf + 1, reply_buf);
	break;

      case 'M':
	cmd_put_memory(packet_buf + 1, reply_buf);
	break;

      case 'c':
	if (cmd_go(0, packet_buf + 1, reply_buf))
	  quit = 1;
	break;

      case 's':
	if (cmd_go(1, packet_buf + 1, reply_buf))
	  quit = 1;
	break;

      case 'Z':
	cmd_set_breakpoint(packet_buf + 1, reply_buf);
	break;

      case 'z':
	cmd_remove_breakpoint(packet_buf + 1, reply_buf);
	break;

      case 'k':
	quit = 1;
	no_reply = 1;
	break;

      default:
	reply_buf[0] = 0;
      }

      if (!no_reply) {
	logmsg(LOG_GDB, 5, "gdb: replying %s\n", reply_buf);
	put_packet(reply_buf);
      }
    }

    close_client_socket();
  }

  close_gdb_socket();
}
