/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>

#include "memory.h"
#include "cpu.h"
#include "hardware.h"
#include "log.h"
#include "lcd.h"

unsigned int tick;

int tick_before(unsigned int a, unsigned int b) {
  return (unsigned int)(b - a) < (1 << 30);
}

static uint32 gpio_read(struct hardware *hardware, uint32 adr) {
  int n = (adr >> 6) & 7;
  int r = adr & 0x3f;
  int i;
  uint32 res;

  switch (r) {
  case 0:
    logmsg(LOG_GPIO, 20, "gpio(%d): read from register %x\n", n, r);
    res = 0;
    for (i = 31; i >= 0; i--) {
      struct gpio_handler *h = &(hardware->gpio_handler[n][i]);
      if (h->get) {
	res = (res << 1) | (*(h->get))(h->data, n, i);
      } else
	res = (res << 1) | 1;
    }
    return res;
  default:
    logmsg(LOG_GPIO, 20, "gpio(%d): read from register %x\n", n, r);
    return ~0;
  }
}

static void gpio_notify_handlers(struct hardware *hardware,
				 int port, uint32 old, uint32 new) {
  int i;
  for (i = 0; i < 32; i++)
    if ((old ^ new) & (1 << i)) {
      struct gpio_handler *h = &(hardware->gpio_handler[port][i]);
      if (h->set)
	(*h->set)(h->data, port, i, new & (1 << i));
    }
      
}

int gpio_get_bit(struct hardware *hardware, int port, int bit) {
  return (hardware->gpio_out[port] >> bit) & 1;
}

void gpio_install_handler(struct hardware *hardware, int port, int bit,
			  void *data,
			  int (*get)(void *, int, int),
			  int (*set)(void *, int, int, int)) {
  hardware->gpio_handler[port][bit].data = data;
  hardware->gpio_handler[port][bit].get = get;
  hardware->gpio_handler[port][bit].set = set;
}

static void gpio_write(struct hardware *hardware, uint32 adr, uint32 val) {
  int n = (adr >> 6) & 7;
  int r = adr & 0x3c;
  uint32 old;

  switch (r) {
  case 0x14:
    old = hardware->gpio_out[n];
    hardware->gpio_out[n] |= val;
    gpio_notify_handlers(hardware,
			 n, old, hardware->gpio_out[n]);
    break;
  case 0x18:
    old = hardware->gpio_out[n];
    hardware->gpio_out[n] &= ~val;
    gpio_notify_handlers(hardware,
			 n, old, hardware->gpio_out[n]);
    break;
  default:
    logmsg(LOG_GPIO, 20, "gpio(%d): write %x to register %x\n", n, val, r);
  }
}

static int flash_ce[4] = {0x20, 0x02, 0x10, 0x08};

static int selected_flash(struct hardware *hardware) {
  int i, r;
  uint32 p5 = hardware->gpio_out[5];
  r = -1;
  for (i = 0; i < 4; i++)
    if (!(p5 & flash_ce[i])) {
      if (r > 0)
	logmsg(LOG_FLASH, 5, "flash: more than one chip selected\n");
      else
	r = i;
    }
  return r;
}

static int flash_get_busy(void *data, int port, int bit) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;

  chip = selected_flash(hardware);
  if (chip == -1)
    return 1;

  flash = hardware->flash[chip];
  if (!flash)
    return 1;

  if (!flash->busy)
    return 1;

  if (tick_before(flash->ready_tick, tick)) {
    flash->busy = 0;
    return 1;
  }

  return 0;
}

static int flash_read_page(struct flash *flash, int page) {
  if (fseek(flash->fp, page * 2112, SEEK_SET) == -1)
    return -1;
  if (fread(flash->reg, 1, 2112, flash->fp) != 2112) return -1;
  return 0;
}

static int flash_write_page(struct flash *flash, int page) {
  if (fseek(flash->fp, page * 2112, SEEK_SET) == -1)
    return -1;
  if (fwrite(flash->reg, 1, 2112, flash->fp) != 2112) return -1;
  return 0;
}

static uint8 flash_read(void *data, uint32 adr) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;
  uint32 r;

  chip = selected_flash(hardware);
  if (chip == -1) {
    logmsg(LOG_FLASH, 5, "flash: no chip selected during read from %x\n", adr);
    return ~0;
  }

  flash = hardware->flash[chip];
  if (!flash)
    return ~0;

  switch (adr & 0xc) {
  case 0:
    if (flash->ptr >= 2112)
      flash->ptr = 0;
    r = flash->reg[flash->ptr];
    flash->ptr++;
    logmsg(LOG_FLASH, 20, "flash(%d): read %x from %x\n", chip, r, adr);
    return r;
  default:
    logmsg(LOG_FLASH, 5, "flash(%d): read from %x\n", chip, adr);
  }
  return ~0;
}

static void flash_read_cmd(struct flash *flash) {
  int i;
  int col = flash->adr[0] | ((flash->adr[1] & 0xf) << 8);
  int row =  flash->adr[2] | (flash->adr[3] << 8)
    | ((flash->adr[4] & 1) << 16);

  logmsg(LOG_FLASH, 10, "flash: op read, address %6x/%3x\n", row, col);

  flash_read_page(flash, row);

  flash->ptr = col;
}

static void flash_page_program(struct flash *flash) {
  int i;
  int row =  flash->adr[2] | (flash->adr[3] << 8)
    | ((flash->adr[4] & 1) << 16);

  logmsg(LOG_FLASH, 10, "flash: op program, address %6x\n", row);

  flash_write_page(flash, row);
}

static void flash_block_erase(struct flash *flash) {
  int row =  flash->adr[0] | (flash->adr[1] << 8)
    | ((flash->adr[2] & 1) << 16);
  int i;

  logmsg(LOG_FLASH, 9, "flash: op block erase, address %6x\n", row);
  row &= ~0x3f;

  for (i = 0; i < 2112; i++)
    flash->reg[i] = 0xff;

  for (i = 0; i < 64; i++)
    flash_write_page(flash, row + i);
}

static void flash_write(void *data, uint32 adr, uint8 val) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip, i;

  chip = selected_flash(hardware);
  if (chip == -1) {
    logmsg(LOG_FLASH, 5, "flash: no chip selected during write %x to %x\n", val, adr);
    return;
  }

  flash = hardware->flash[chip];
  if (!flash)
    return;

  switch (adr & 0xc) {
  case 0:
    if (flash->ptr >= 2112)
      flash->ptr = 0;
    flash->reg[flash->ptr] = val;
    flash->ptr++;
    logmsg(LOG_FLASH, 20, "flash(%d): write %x to %x\n", chip, val, adr);
    break;
  case 4:
    flash->adr_ptr = 0;
    logmsg(LOG_FLASH, 19, "flash(%d): command %x\n", chip, val);

    switch (val) {
    case 0x90:
      flash->cmd = val;
      logmsg(LOG_FLASH, 9, "flash(%d): read id\n", chip);
      flash->reg[0] = 0xec;
      flash->reg[1] = 0xda; /* 0xaa chyba nie jest rozpoznawane */
      flash->reg[2] = 0xff;
      flash->reg[3] = 0x15;
      flash->ptr = 0;
      break;
    case 0x70:
      flash->cmd = val;
      flash->reg[0] = 0xe0;
      flash->ptr = 0;
      break;
    case 0x00:
    case 0x05:
    case 0x60:
    case 0x80:
      flash->cmd = val;
      break;
    case 0x10:
      flash_page_program(flash);
      break;
    case 0x30:
      flash_read_cmd(flash);
      break;
    case 0xd0:
      flash_block_erase(flash);
      break;
    case 0xe0:
      i = flash->adr[0] | ((flash->adr[1] & 0xf) << 8);
      flash->ptr = i;
      logmsg(LOG_FLASH, 10, "flash: read random, column %3x\n", i);
      break;
    default:
      break;
    }
    break;
  case 8:
    logmsg(LOG_FLASH, 19, "flash(%d): address %x\n", chip, val);
    if (flash->adr_ptr >= 5)
      flash->adr_ptr = 0;

    flash->adr[flash->adr_ptr] = val;
    flash->adr_ptr++;

    if (flash->cmd == 0x80 && flash->adr_ptr == 5) {
      int row =  flash->adr[2] | (flash->adr[3] << 8)
	| ((flash->adr[4] & 1) << 16);
      flash->ptr = flash->adr[0] | ((flash->adr[1] & 0xf) << 8);
      flash_read_page(flash, row);
    }

    break;
  default:
    logmsg(LOG_FLASH, 5, "flash(%d): write %x to %x\n", chip, val, adr);
  }

}

void free_flash(struct flash *flash) {
  if (flash) {
    if (flash->fp)
      fclose(flash->fp);
    free(flash);
  }
}

static uint32 prog_flash_read(void *data, uint32 adr) {
  struct prog_flash *flash = data;
  logmsg(LOG_FW_FLASH, 10, "prog_flash: read from %x\n", adr);
  adr = adr & 0xffffe;

  switch (flash->state) {
  case 0:
    return flash->content[adr] | (flash->content[adr + 1] << 8);

  case 0x90:
    switch (adr) {
    case 0:
      return 0xbf;
    case 2:
      return 0x2781;
    default:
      return 0xffff;
    }
  default:
    return 0xffff;
  }
}

static uint8 prog_flash_read_byte(void *data, uint32 adr) {
  uint8 res;
  if (adr & 1)
    res = prog_flash_read(data, adr & ~1) >> 8;
  else
    res = prog_flash_read(data, adr) & 0xff;
  logmsg(LOG_FW_FLASH, 10, "prog_flash: read byte %x\n", res);
  return res;
}

static int prog_flash_update(struct prog_flash *flash, int start, int size) {
  if (fseek(flash->fp, start, SEEK_SET))
    return -1;
  fwrite(flash->content + start, 1, size, flash->fp);
  return 0;
}

static void prog_flash_sector_erase(struct prog_flash *flash, int sector) {
  sector = sector & 0xff;
  memset(flash->content + sector * 4096, 0xff, 4096);
  prog_flash_update(flash, sector * 4096, 4096);
}

static void prog_flash_block_erase(struct prog_flash *flash, int block) {
  block = block & 0xf;
  memset(flash->content + block * 64 * 1024, 0xff, 64 * 1024);
  prog_flash_update(flash, block * 64 * 1024, 64 * 1024);
}

static void prog_flash_chip_erase(struct prog_flash *flash) {
  memset(flash->content, 0xff, 1024 * 1024);
  prog_flash_update(flash, 0, 1024 * 1024);
}

static void prog_flash_write(void *data, uint32 adr, uint32 val) {
  struct prog_flash *flash = data;
  int adr15 = (adr & 0xffff) >> 1;
  int val8 = val & 0xff;

  logmsg(LOG_FW_FLASH, 10, "prog_flash: write %x to %x\n", val, adr);

  switch (flash->state) {
  case 0:
    if (adr15 == 0x5555 && val8 == 0xaa)
      flash->state = 1;
    break;

  case 1:
    if (adr15 == 0x2aaa && val8 == 0x55)
      flash->state = 2;
    else
      flash->state = 0;
    break;

  case 2:
    if (adr15 == 0x5555) {
      switch (val8) {
      case 0xa0:
      case 0x80:
      case 0x90:
      case 0x98:
	flash->state = val & 0xff;
	break;
      default:
	flash->state = 0;
	break;
      }
    } else
      flash->state = 0;
    break;

  case 0x80:
    /* erase mode */
    if (adr15 == 0x5555 && val8 == 0xaa)
      flash->state = 0x81;
    else
      flash->state = 0;
    break;

  case 0x81:
    if (adr15 == 0x2aaa && val8 == 0x55)
      flash->state = 0x82;
    else
      flash->state = 0;
    break;

  case 0x82:
    switch (val8) {
    case 0x30:
      prog_flash_sector_erase(flash, adr >> 12);
      flash->state = 0;
      break;
    case 0x50:
      prog_flash_block_erase(flash, adr >> 16);
      flash->state = 0;
      break;
    case 0x10:
      if (adr15 == 0x5555)
	prog_flash_chip_erase(flash);
      flash->state = 0;
      break;
    default:
      flash->state = 0;
      break;
    }
    break;
  case 0x90:
    /* software id mode */
    if (adr15 == 0x5555 && val8 == 0xf0)
      flash->state = 0;
    break;

  case 0x98:
    /* TODO: CFI query */
    flash->state = 0;
    break;

  case 0xa0:
    adr = adr & 0xffffe;
    flash->content[adr] = val & 0xff;
    flash->content[adr + 1] = val >> 8;
    prog_flash_update(flash, adr, 2);
    flash->state = 0;
    break;

  default:
    flash->state = 0;
    break;
  }
}

static uint32 buf_803;
static uint32 read_803(struct hardware *hardware, uint32 adr) {
  logmsg(LOG_MEMORY, 5, "read from %x\n", adr);
  switch (adr) {
  default:
    return buf_803;
  }
}

static void write_803(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_MEMORY, 5, "write %x to %x\n", val, adr);
  buf_803 = val;
}

static uint32 adc_read(struct hardware *hardware, uint32 adr) {
  switch (adr) {
  case 0: /* most of buttons */
    if (ui_get_key(2)) return 0x280;
    if (ui_get_key(3)) return 0x300;
    if (ui_get_key(4)) return 0x100;
    if (ui_get_key(5)) return 0x20;
    if (ui_get_key(6)) return 0x60;
    if (ui_get_key(7)) return 0x1a0;
    if (ui_get_key(8)) return 0x200;
    return 0x3ff;
  case 4: /* battery */
    return 450;
  case 8:
    if (ui_get_key(1)) return 0;
    return 0x3ff;
  default:
    logmsg(LOG_ADC, 5, "adc: read from %x\n", adr);
  }
  return ~0;
}

static void adc_write(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_ADC, 5, "adc: write %x to %x\n", val, adr);
}

static uint32 i2c_read(struct hardware *hardware, uint32 adr) {
  switch (adr & 0xff) {
  case 4:
    logmsg(LOG_I2C, 5, "i2c: status read\n", adr);
    return 0;
  default:
    logmsg(LOG_I2C, 5, "i2c: read from %x\n", adr);
  }
  return ~0;
}

static void i2c_write(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_I2C, 5, "i2c: write %x to %x\n", val, adr);
}

static uint32 hardware_read_word(void *data, uint32 adr) {
  struct hardware *hardware = data;

  if (adr == 0x80105000) {
    logmsg(LOG_MEMORY, 5, "read from %8x (returning 0)\n", adr);
    return 0;
  }
  switch ((adr >> 12) & 0xfff) {
  case 3:
    return gpio_read(data, adr);
  case 0x104:
    return lcd_read(hardware->lcd, adr);
  case 0x300:
    return read_803(data, adr);
  default:
    if ((adr >> 8) == 0x800024) 
      return adc_read(hardware, adr & 0xff);
    else if ((adr >> 8) == 0x800208) 
      return i2c_read(hardware, adr & 0xff);

    logmsg(LOG_MEMORY, 10, "read from %8x\n", adr);
    return 0xffffffff;
  }
}

static void hardware_write_word(void *data, uint32 adr, uint32 val) {
  struct hardware *hardware = data;

  switch ((adr >> 12) & 0xfff) {
  case 3:
    gpio_write(data, adr, val);
    break;
  case 0x104:
    lcd_write(hardware->lcd, adr, val);
    break;
  case 0x300:
    write_803(data, adr, val);
    break;
  default:
    if ((adr >> 8) == 0x800024) 
      adc_write(hardware, adr & 0xff, val);
    else if ((adr >> 8) == 0x800208)
      i2c_write(hardware, adr & 0xff, val);
    else
      logmsg(LOG_MEMORY, 10, "write %8x to %8x (pc=%8x)\n", val, adr, cpu->reg[15] - 4);
  }
}

void install_hardware(struct hardware *hardware,
		      struct memory *memory) {
  set_area(memory, 0x80000000, 0x803fffff,
	   read_void, write_void,
	   read_half_word_void, write_half_word_void,
	   hardware_read_word, hardware_write_word, hardware);
  set_area(memory, 0x20000000, 0x200fffff,
	   prog_flash_read_byte, write_void,
	   prog_flash_read, prog_flash_write,
	   read_word_void, write_word_void, hardware->prog_flash);
  set_area(memory, 0x28000000, 0x2800ffff,
	   flash_read, flash_write,
	   read_half_word_void, write_half_word_void,
	   read_word_void, write_word_void, hardware);
}

struct prog_flash *create_prog_flash(char *fname) {
  struct prog_flash *flash;

  flash = xmalloc(sizeof(struct prog_flash));
  if (!flash)
    return NULL;

  flash->state = 0;

  flash->content = xmalloc(1024 * 1024);
  if (!flash->content) {
    free(flash);
    return NULL;
  }
  memset(flash->content, 0xff, 1024 * 1024);

  flash->fp = fopen(fname, "r+");
  if (!flash->fp) {
    error_sys("%s: ", fname);
    free(flash->content);
    free(flash);
    return NULL;
  }

  fread(flash->content, 1, 1024 * 1024, flash->fp);
  return flash;
}

void free_prog_flash(struct prog_flash *flash) {
  if (flash) {
    if (flash->fp)
      fclose(flash->fp);
    if (flash->content)
      free(flash->content);
    free(flash);
  }
}

void hardware_reset(struct hardware *hardware) {
  int i;
  tick = 0;
  for (i = 0; i < 8; i++)
    hardware->gpio_out[i] = ~0;
}

static int usb_get_connected(void *data, int port, int bit) {
  return 0;
}

static int get_hold(void *data, int port, int bit) {
  return 1;
}

static int get_port6b0(void *data, int port, int bit) {
  return 0;
}

struct hardware *create_hardware(void) {
  int i, j;
  struct hardware *hardware = xmalloc(sizeof(struct hardware));
  if (!hardware) return NULL;

  hardware->prog_flash = create_prog_flash(prog_flash_file);
  if (!hardware->prog_flash)
    goto no_prog_flash;

  for (i = 0; i < 4; i++)
    hardware->flash[i] = NULL;

  hardware->flash[0] = xmalloc(sizeof(struct flash));
  if (!hardware->flash[0])
    goto no_flash;

  hardware->flash[0]->ptr = 0;
  hardware->flash[0]->fp = fopen(flash_file, "r+b");
  hardware->flash[0]->busy = 0;
  if (!hardware->flash[0]->fp) {
    error_sys("%s: ", flash_file);
    goto no_fp;
  }

  for (i = 0; i < 8; i++)
    for (j = 0; j < 32; j++) {
      hardware->gpio_handler[i][j].get = NULL;
      hardware->gpio_handler[i][j].set = NULL;
    }

  gpio_install_handler(hardware, 6, 3,
		       hardware, flash_get_busy, NULL);

  gpio_install_handler(hardware, 7, 0,
		       hardware, usb_get_connected, NULL);

  gpio_install_handler(hardware, 5, 2,
		       hardware, get_hold, NULL);

  gpio_install_handler(hardware, 6, 0,
		       hardware, get_port6b0, NULL);

  hardware->lcd = create_lcd();
  if (!hardware->lcd)
    goto no_lcd;

  return hardware;

 no_lcd:
  fclose(hardware->flash[0]->fp);
 no_fp:
  free(hardware->flash[0]);
 no_flash:
  free_prog_flash(hardware->prog_flash);
 no_prog_flash:
  free(hardware);
 err:
  return NULL;

}
