/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdlib.h>

#include "log.h"
#include "lcd.h"

uint32 lcd_read(struct lcd *lcd, uint32 adr) {
  int r;

  switch (adr & 0x3f) {
  case 8:
    logmsg(LOG_LCD, 15, "lcd: status read\n");
    r = 3;
    /*
    if (!(lcd->flags & LCDFLG_ADC))
      r |= 0x40;
    if (!(lcd->flags & LCDFLG_DISPLAY_ON))
      r |= 0x20;
    */
    return r;

  default:
    logmsg(LOG_LCD, 5, "lcd: read from %x\n", adr);
    return ~0;
  }
}

static void lcd_write_control(struct lcd *lcd, uint32 val) {

  if (lcd->flags & LCDFLG_REF_VOLTAGE_SEL) {
    logmsg(LOG_LCD, 10, "lcd: reference voltage select %x\n", val & 0x3f);
    lcd->flags &= ~LCDFLG_REF_VOLTAGE_SEL;

  } else if (lcd->flags & LCDFLG_SET_STATIC_IND_MODE) {
    logmsg(LOG_LCD, 10, "lcd: set static indicator mode %x\n", val & 3);
    lcd->flags &= ~LCDFLG_SET_STATIC_IND_MODE;

  } else {

    if ((val & 0xfe) == 0xae) {
      logmsg(LOG_LCD, 10, "lcd: display on (%x)\n", val & 1);
      if (val & 1)
	lcd->flags |= LCDFLG_DISPLAY_ON;
      else
	lcd->flags &= ~LCDFLG_DISPLAY_ON;

    } else if ((val & 0xc0) == 0x40) {
      logmsg(LOG_LCD, 10, "lcd: initial display line %x\n", val & 0x3f);

    } else if (val == 0x81) {
      lcd->flags |= LCDFLG_REF_VOLTAGE_SEL;

    } else if ((val & 0xf0) == 0xb0) {
      logmsg(LOG_LCD, 10, "lcd: set page address %x\n", val & 0xf);
      lcd->page = val & 0xf;

    } else if ((val & 0xf0) == 0x10) {
      logmsg(LOG_LCD, 10, "lcd: set column address MSB %x\n", val & 0xf);
      lcd->col = (lcd->col & 0xf) | ((val & 0xf) << 4);

    } else if ((val & 0xf0) == 0x00) {
      logmsg(LOG_LCD, 10, "lcd: set column address LSB %x\n", val & 0xf);
      lcd->col = (lcd->col & 0xf0) | (val & 0xf);

    } else if ((val & 0xfe) == 0xa0) {
      logmsg(LOG_LCD, 10, "lcd: set ADC %x\n", val & 1);
      if (val & 1)
	lcd->flags |= LCDFLG_ADC;
      else
	lcd->flags &= ~LCDFLG_ADC;

    } else if ((val & 0xfe) == 0xa6) {
      logmsg(LOG_LCD, 10, "lcd: reverse display %x\n", val & 1);

    } else if ((val & 0xfe) == 0xa4) {
      logmsg(LOG_LCD, 10, "lcd: entire display on %x\n", val & 1);

    } else if ((val & 0xfe) == 0xa2) {
      logmsg(LOG_LCD, 10, "lcd: bias %x\n", val & 1);

    } else if (val == 0xe0) {
      logmsg(LOG_LCD, 10, "lcd: set modify-read\n");

    } else if (val == 0xee) {
      logmsg(LOG_LCD, 10, "lcd: reset modify-read\n");

    } else if (val == 0xe2) {
      logmsg(LOG_LCD, 10, "lcd: reset\n");

    } else if ((val & 0xf0) == 0xc0) {
      logmsg(LOG_LCD, 10, "lcd: shl select %x\n", (val >> 3) & 1);

    } else if ((val & 0xf8) == 0x28) {
      logmsg(LOG_LCD, 10, "lcd: power control %x\n", val & 7);

    } else if ((val & 0xf8) == 0x20) {
      logmsg(LOG_LCD, 10, "lcd: select regulator resistor %x\n", val & 7);

    } else if ((val & 0xfe) == 0xac) {
      if (val & 1)
	lcd->flags |= LCDFLG_SET_STATIC_IND_MODE;

    } else if (val == 0xe3) {
      logmsg(LOG_LCD, 10, "lcd: nop\n");

    } else if ((val & 0xf0) == 0xf0 || (val & 0xf0) == 0x90) {
      logmsg(LOG_LCD, 5, "lcd: test instruction\n");

    } else
      logmsg(LOG_LCD, 5, "lcd: write control %x\n", val);
  }
}

static void lcd_write_data(struct lcd *lcd, uint32 val) {
  if (lcd->page < 9 && lcd->col < 132) {
    logmsg(LOG_LCD, 10, "lcd: write data %x\n", val);
    lcd->ddram[132 * lcd->page + lcd->col] = val;
    lcd->col++;
  } else
    logmsg(LOG_LCD, 5, "lcd: out of display write data %x\n", val);
}

void lcd_write(struct lcd *lcd, uint32 adr, uint32 val) {
  switch (adr & 0x3f) {
  case 0x20:
    lcd_write_control(lcd, val);
    break;
  case 0x30:
    lcd_write_data(lcd, val);
    break;
  default:
    logmsg(LOG_LCD, 5, "lcd: write %x to %x\n", val, adr);
    break;
  }
}

struct lcd *create_lcd(void) {
  struct lcd *lcd;
  lcd = xmalloc(sizeof(struct lcd));
  if (!lcd)
    return NULL;

  memset(lcd->ddram, 0, sizeof(lcd->ddram));

  lcd->flags = 0;
  return lcd;
}

void free_lcd(struct lcd *lcd) {
  if (lcd) free(lcd);
}

int lcd_get_pixel(struct lcd *lcd, int x, int y) {
  return (lcd->ddram[132 * (y / 8) + x] >> (y % 8)) & 1;
}

static print_small_chars[5] = {'.', 'o', 'o', 'O', '#'};

void lcd_print_small(struct lcd *lcd, FILE *fp) {
  int x, y;
  for (y = 0; y < 32; y++) {
    for (x = 0; x < 64; x++) {
      int c = 0;
      if (lcd_get_pixel(lcd, x * 2, y * 2)) c++;
      if (lcd_get_pixel(lcd, x * 2 + 1, y * 2)) c++;
      if (lcd_get_pixel(lcd, x * 2, y * 2 + 1)) c++;
      if (lcd_get_pixel(lcd, x * 2 + 1, y * 2 + 1)) c++;
      fputc(print_small_chars[c], fp);
    }
    fputc('\n', fp);
  }
}
